/*
 * Copyright (C) 2002-2003  Pekka Enberg <Pekka.Enberg@cs.helsinki.fi>
 *
 * Distributed under the terms of the GNU General Public License
 * version 2 or later.
 */
#include <assert.h>
#include <curses.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "ansi-esc.h"
#include "bin-file.h"
#include "colors.h"
#include "editor-context.h"
#include "edit-buffer.h"
#include "error.h"
#include "screen.h"

/* Curses type KEY_xxx check if META key was used for character.  */
#define KEY_META(ch) (0x1000 | ch)

/* ALT and ESC are both the same META key for terminals.  */
#define META_KEY_CODE 0x1B

int
get_char(void)
{
	int ch = getch();
	if (ch == META_KEY_CODE) {
		/* After META comes the actual key we're interested in.  */
		ch = KEY_META(getch());
	}
	return ch;
}

int
get_printable_char(int ch)
{
	/* TODO: Function keys should go trough lookup table.  */
	switch (ch) {
	case KEY_F(1):
		return 176;
	case KEY_F(2):
		return 177;
	case KEY_F(3):
		return 178;
	case KEY_F(4):
		return 219;
	case KEY_F(5):
		return 220;
	case KEY_F(6):
		return 223;
	case KEY_F(7):
		return 221;
	case KEY_F(8):
		return 222;
	case KEY_F(10):
		return 254;
	}
	return ch;
}

/*
 * Commands issued in editor
 */

void
cmd_resize(void)
{
	/* Currently we don't support resizing the terminal while
	   running.  */
	error("Terminal was resized.");
}

void
cmd_move_up(struct edit_buffer *buf, struct screen *scr)
{
	if (scr->cursor_y > 0)
		scr->cursor_y--;
	else if (buf->start_y > 0)
		buf->start_y--;
}

void
cmd_move_down(struct edit_buffer *buf, struct screen *scr)
{
	if (scr->cursor_y < (scr->height - 1))
		scr->cursor_y++;
	else if (buf->start_y + scr->height < buf->height)
		buf->start_y++;
}

void
cmd_move_left(struct edit_buffer *buf, struct screen *scr)
{
	if (scr->cursor_x > 0)
		scr->cursor_x--;
	else if (buf->start_x > 0)
		buf->start_x--;
}

void
cmd_move_right(struct edit_buffer *buf, struct screen *scr)
{
	if (scr->cursor_x < (scr->width - 1))
		scr->cursor_x++;
	else if (buf->start_x < (buf->width - scr->width))
		buf->start_x++;
}

void
cmd_move_to_end(struct edit_buffer *buf, struct screen *scr)
{
	scr->cursor_x = (scr->width - 1);
	buf->start_x = buf->width - scr->width;
}

void
cmd_move_to_start(struct edit_buffer *buf, struct screen *scr)
{
	scr->cursor_x = 0;
	buf->start_x = 0;
}

void
cmd_print_char(struct edit_buffer *buf, struct screen *scr,
	       struct editor_context *ctx, int ch)
{
	edit_buffer_put(buf,
			buf->start_x + scr->cursor_x,
			buf->start_y + scr->cursor_y,
			(COLOR_ATTR(ctx->fg_color, ctx->bg_color) << 8)
			| get_printable_char(ch));
}

void
cmd_move_page_up(struct edit_buffer *buf, struct screen *scr)
{
	if (buf->start_y > scr->height) {
		buf->start_y -= scr->height;
	} else {
		buf->start_y = 0;
		scr->cursor_y = 0;
	}
	screen_redraw();
}

void
cmd_move_page_down(struct edit_buffer *buf, struct screen *scr)
{
	if (buf->start_y + (scr->height * 2) < buf->height) {
		buf->start_y += scr->height;
	} else {
		buf->start_y = buf->height - scr->height;
		scr->cursor_y = (scr->height - 1);
	}
	screen_redraw();
}

#define MAX_BG_COLOR 7
#define MAX_FG_COLOR 15

#define DECREASE_AND_WRAP(c, max) (c >   0 ? c - 1 : max)
#define INCREASE_AND_WRAP(c, max) (c < max ? c + 1 :   0)

void
cmd_next_fg_color(struct editor_context *ctx)
{
	ctx->fg_color = INCREASE_AND_WRAP(ctx->fg_color, MAX_FG_COLOR);
}

void
cmd_prev_fg_color(struct editor_context *ctx)
{
	ctx->fg_color = DECREASE_AND_WRAP(ctx->fg_color, MAX_FG_COLOR);
}

void
cmd_next_bg_color(struct editor_context *ctx)
{
	ctx->bg_color = INCREASE_AND_WRAP(ctx->bg_color, MAX_BG_COLOR);
}

void
cmd_prev_bg_color(struct editor_context *ctx)
{
	ctx->bg_color = DECREASE_AND_WRAP(ctx->bg_color, MAX_BG_COLOR);
}

/*
 * Main editor loop
 */

static void edit_loop(struct edit_buffer *buf)
{
	struct editor_context _ctx = {
		.fg_color = 0x07,
		.bg_color = 0x00
	};
	struct editor_context *ctx = &_ctx;
	struct screen *scr = screen_init();
	bool quit = false;
	while (!quit) {
		screen_draw_edit_buffer(scr, buf);
		screen_print_status(buf, scr, ctx);
		move(scr->cursor_y, scr->cursor_x);

		int ch = get_char();
		if (ch == ERR) {
			error("getch() returned ERR");
		}
		switch (ch) {
		case KEY_META('x'):
		case KEY_META('X'):
			quit = true;
			break;
		case KEY_RESIZE:
			cmd_resize();
			break;
		case KEY_UP:
			cmd_move_up(buf, scr);
			break;
		case KEY_DOWN:
			cmd_move_down(buf, scr);
			break;
		case KEY_LEFT:
			cmd_move_left(buf, scr);
			break;
		case KEY_RIGHT:
			cmd_move_right(buf, scr);
			break;
		case KEY_END:
			cmd_move_to_end(buf, scr);
			break;
		case KEY_HOME:
			cmd_move_to_start(buf, scr);
			break;
		case KEY_PPAGE:
			cmd_move_page_up(buf, scr);
			break;
		case KEY_NPAGE:
			cmd_move_page_down(buf, scr);
			break;
		case KEY_META(KEY_UP):
			cmd_next_bg_color(ctx);
			break;
		case KEY_META(KEY_DOWN):
			cmd_prev_bg_color(ctx);
			break;
		case KEY_META(KEY_RIGHT):
			cmd_next_fg_color(ctx);
			break;
		case KEY_META(KEY_LEFT):
			cmd_prev_fg_color(ctx);
			break;
		case KEY_BACKSPACE:
			cmd_move_left(buf, scr);
			cmd_print_char(buf, scr, ctx, ' ');
			break;
		default:
			cmd_print_char(buf, scr, ctx, ch);
			cmd_move_right(buf, scr);
		}
	}
	screen_release(scr);
}

static void usage(char * argv[])
{
	printf("usage: %s [-c <columns> -r <rows>] [filename]\n", argv[0]);
}

int main(int argc, char *argv[])
{
	unsigned long edit_buffer_cols = 160;
	unsigned long edit_buffer_rows = 1000;

	for (;;) {
		int arg_index = getopt(argc, argv, "hc:r:");
		if (arg_index == -1) {
			break;
		}
		switch (arg_index) {
		case 'h':
			usage(argv);
			return EXIT_SUCCESS;
		case 'c':
			edit_buffer_cols = atoi(optarg);
			break;
		case 'r':
			edit_buffer_rows = atoi(optarg);
			break;
		default:
			usage(argv);
			return EXIT_FAILURE;
		}
	}
	struct edit_buffer *buf = edit_buffer_create(edit_buffer_cols, edit_buffer_rows);
	edit_buffer_clear(buf);
	if (argv[optind] != NULL) {
		FILE *input = fopen(argv[optind], "r");
		if (input) {
			if (bin_file_check(argv[optind]))
				bin_file_translate(input, buf, buf->width);
			else
				ansi_esc_translate(input, buf);
			fclose(input);
		}
	}
	edit_loop(buf);
	edit_buffer_release(buf);
	return EXIT_SUCCESS;
}
